package org.jvnet.jaxbcommons.hashcode.addon.generator;

import org.jvnet.jaxbcommons.addon.Ignoring;
import org.jvnet.jaxbcommons.addon.generator.AbstractMethodStrategy;
import org.jvnet.jaxbcommons.util.AccessorUtils;
import org.jvnet.jaxbcommons.util.FieldUtils;

import com.sun.codemodel.JBlock;
import com.sun.codemodel.JCodeModel;
import com.sun.codemodel.JConditional;
import com.sun.codemodel.JDefinedClass;
import com.sun.codemodel.JExpr;
import com.sun.codemodel.JExpression;
import com.sun.codemodel.JForLoop;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JMod;
import com.sun.codemodel.JOp;
import com.sun.codemodel.JType;
import com.sun.codemodel.JVar;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.grammar.ClassItem;
import com.sun.tools.xjc.grammar.FieldItem;
import com.sun.tools.xjc.grammar.FieldUse;

public class HashCodeStrategy extends AbstractMethodStrategy {

	private final Ignoring ignoring;

	public HashCodeStrategy(Ignoring ignoring) {
		this.ignoring = ignoring;
	}

	public JMethod generate(ClassContext classContext, FieldItem fieldItem,
			JDefinedClass theClass) {
		final ClassItem classItem = classContext.target;
		if (!ignoring.isIgnored(classItem)) {
			final JCodeModel codeModel = getCodeModel(classContext);
			final JMethod hashCode = classContext.implClass.method(JMod.PUBLIC,
					codeModel.INT, "hashCode");
			final JVar hash = hashCode.body().decl(codeModel.INT, "hash",
					JExpr.lit(7));

			for (ClassItem item = classItem; item != null; item = item
					.getSuperClass()) {
				final ClassContext context = classContext.parent
						.getClassContext(item);
				generateFieldsHashCode(context, hashCode.body(), hash);
			}

			hashCode.body()._return(hash);
			return hashCode;
		} else {
			return null;
		}
	}

	/**
	 * Generates fields hashcode instructions.
	 * 
	 * @param classContext
	 *            class context.
	 * @param methodBody
	 *            method body.
	 * @param hash
	 *            hash variable.
	 */
	protected void generateFieldsHashCode(final ClassContext classContext,
			final JBlock methodBody, final JVar hash) {
		final JCodeModel codeModel = getCodeModel(classContext);
		final FieldUse[] fieldUses = classContext.target.getDeclaredFieldUses();
		for (int index = 0; index < fieldUses.length; index++) {
			final FieldUse fieldUse = fieldUses[index];
			final FieldItem fieldItem = FieldUtils.getFieldItem(fieldUse);
			if (!ignoring.isIgnored(fieldItem)) {
				final JMethod getter = AccessorUtils.get(classContext,
						fieldItem);
				final JMethod isSetter = AccessorUtils.isSet(classContext,
						fieldItem);

				if (getter != null) {
					final JBlock block = methodBody.block();
					if (isSetter == null) {
						final JVar value = block.decl(getter.type(), "value",
								JExpr._this().invoke(getter));
						final JExpression isValueSet = JOp.eq(JExpr._null(),
								value);
						appendHashCode(codeModel, block, hash, getter.type(),
								value, isValueSet);
					} else {
						appendHashCode(codeModel, block, hash, getter.type(),
								JExpr._this().invoke(getter), JExpr._this()
										.invoke(isSetter).not());
					}
				}
			}
		}
	}

	private void appendHashCode(final JCodeModel codeModel, final JBlock block,
			final JVar hash, final JType type, final JExpression value,
			final JExpression isValueSet) {
		if (type.isArray()) {

			final JConditional _if = block._if(JExpr._null().eq(value));
			_if._then().assign(hash, JOp.mul(JExpr.lit(31), hash));
			final JForLoop _for = _if._else()._for();
			final JVar i = _for.init(codeModel.INT, "i", JExpr.lit(0));
			_for.test(JOp.lt(i, value.ref("length")));
			_for.update(JOp.incr(i));
			final JVar valueComponent = _for.body().decl(type.elementType(),
					"valueComponent", value.component(i));
			_for.body().assign(
					hash,
					JOp.plus(
							JOp.mul(JExpr.lit(31), hash),
							valueHashCode(codeModel, type.elementType(),
									valueComponent,
									JOp.eq(JExpr._null(), valueComponent))));
		} else {
			block.assign(
					hash,
					JOp.plus(JOp.mul(JExpr.lit(31), hash),
							valueHashCode(codeModel, type, value, isValueSet)));
		}
	}

	/**
	 * Generates hash code expression for value.
	 * 
	 * @param value
	 *            value.
	 * @return Hash code expression for the value.
	 */
	protected JExpression valueHashCode(final JCodeModel codeModel,
			final JType type, final JExpression value,
			final JExpression isValueSet) {

		if ("byte".equals(type.fullName())) {
			return JExpr.cast(codeModel.INT, value);
		} else if ("short".equals(type.fullName())) {
			return JExpr.cast(codeModel.INT, value);
		} else if ("boolean".equals(type.fullName())) {
			return JOp.cond(value, JExpr.lit(1), JExpr.lit(0));
		} else if ("char".equals(type.fullName())) {
			return JExpr.cast(codeModel.INT, value);
		} else if ("int".equals(type.fullName())) {
			return value;
		} else if ("long".equals(type.fullName())) {
			return JExpr.cast(codeModel.INT,
					JOp.xor(value, JOp.shrz(value, JExpr.lit(32))));
		} else if ("float".equals(type.fullName())) {
			return codeModel.ref(Float.class).staticInvoke("floatToIntBits")
					.arg(value);
		} else if ("double".equals(type.fullName())) {
			return JExpr.cast(
					codeModel.INT,
					JOp.xor(codeModel.ref(Double.class)
							.staticInvoke("doubleToLongBits").arg(value),
							JOp.shrz(
									codeModel.ref(Double.class)
											.staticInvoke("doubleToLongBits")
											.arg(value), JExpr.lit(32))));
		} else {
			return JOp.cond(isValueSet, JExpr.lit(0), value.invoke("hashCode"));
		}
	}

}
